//
//
// HMC Lion knock board prototype
//
// set lfuse to 0x5E for 20 MHz xtal
//
// Lincoln Craven-Brightman
// 19/11/05
//

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define go_signal_pins PINA
#define go_signal_port PORTA
#define out_signal_port PORTA
#define out_direction DDRA
#define go_signal_pin (1 << PA2)
#define out_signal_pin (1 << PA3)

#define led_R_direction DDRA
#define led_G_direction DDRB
#define led_R_pin (1 << PA7)
#define led_G_pin (1 << PB2)
#define led_R_port PORTA
#define led_G_port PORTB

#define piezo_pin (1 << PA1)
#define piezo_port PORTA
#define piezo_pins PINA

#define max_knocks 9
#define block_length 40 // delay block length in ms
#define min_delays 3 // min number of delay blocks between knocks


void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
}


int block_measure(volatile unsigned char *port, unsigned char pin, unsigned int *lp_i, unsigned int n_loops, unsigned int *high) {
   *high = 0;
   for (*lp_i = 0; *lp_i < n_loops; (*lp_i)++) {
      if (*high) {
         _delay_us(20);
      }
      else {
         // initiate conversion
         ADCSRA |= (1 << ADSC);
         // wait for completion (should be 13us)
         while (ADCSRA & (1 << ADSC))
            ;
         // check if result is over threshold (~3.5V)
         if (ADCH >= 3) {
            *high = 1;
         }
         _delay_us(7);
      }
   }
   return *high;
}

void put_knocks(volatile unsigned char *port, unsigned char pin, char knocks) {
   // put knocks after delay
   _delay_ms(100);
   put_char(port, pin, knocks);
   char_delay();
}


int main(void) {
   //
   // main
   //
   static unsigned int loop_var;
   static char knocks = 0;
   static unsigned int delays;
   static unsigned int high;
   static int sent = 0;
   //
   // set clock divider to /1
   //
   CLKPR = (1 << CLKPCE);
   CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
   //
   // initialize output pins
   //
   set(out_signal_port, out_signal_pin);
   set(go_signal_port, go_signal_pin); // pull go signal high
   output(led_G_direction, led_G_pin);
   output(led_R_direction, led_R_pin);
   output(out_direction, out_signal_pin);

   //
   // init A/D
   //
   ADMUX = (0 << REFS1) | (0 << REFS0)  // Use VCC for reference voltage
         | (0 << MUX5) | (0 << MUX4) | (0 << MUX3)
         | (0 << MUX2) | (0 << MUX1) | (1 << MUX0);  // 000001 = Use ADC1 (PA1)
   ADCSRA = (1 << ADEN) // enable
      | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128

   //
   // boot delay
   //
   set(led_R_port, led_R_pin);
   _delay_ms(500);
   clear(led_R_port, led_R_pin);
   //
   // main loop
   //
   while (1) {
      if (!pin_test(go_signal_pins, go_signal_pin)) {
         sent = 0;
         set(led_G_port, led_G_pin);
         // If go_signal:
         // measure [block length] of piezo
         if (block_measure(&piezo_pins, piezo_pin, &loop_var, ((block_length / 20) * 1000), &high)) {
            // If high:
            // if delays long enough, add 1 to knocks (cap at max_knocks)
            // reset delays
            if (delays >= min_delays && knocks < max_knocks) {
               knocks ++;
               set(led_R_port, led_R_pin);
               _delay_ms(10);
               clear(led_R_port, led_R_pin);
               delays = 0;
            }
            
         }
         // add 1 to delays, without overflow
         if (delays <= min_delays) delays ++;
      }
      else if (!sent) {
         // else:
         // put knocks as output
         // reset
         //
         put_knocks(&out_signal_port, out_signal_pin, knocks);
         knocks = 0;
         delays = 0;
         sent = 1;
      }
      else {
         clear(led_G_port, led_G_pin);
      }
   }
}
