//
//
// HMC Lion load cell board prototype
//
// set lfuse to 0x5E for 20 MHz xtal
//
// Lincoln Craven-Brightman
// 19/12/09
//

#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#define output(directions,pin) (directions |= pin) // set port direction for output
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 8.5 // bit delay for 115200 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay
#define char_delay() _delay_ms(10) // char delay

#define input_pins PINA
#define input_port PORTA
#define green_pin (1 << PA2)
#define white_pin (1 << PA1)

#define led_pin (1 << PA3)
#define led_port PORTA
#define led_direction DDRA

#define digit_pin (1 << PA7)
#define go_pin (1 << PB2)
#define digit_port PORTA
#define go_port PORTB
#define digit_direction DDRA
#define go_direction DDRB

#define serial_pin_out (1 << PA5)
#define serial_port PORTA
#define serial_direction DDRA

#define threshold 2
#define nsamples 100 // number of samples to accumulate

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
}


int main(void) {
	//
	// main
	//
	static char negative = 0;
	static uint16_t count;
   	static uint32_t accum;
	//
	// set clock divider to /1
	//
	CLKPR = (1 << CLKPCE);
	CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);
	//
	// initialize pins
	//
	set(digit_port, digit_pin);
	set(serial_port, serial_pin_out);
	set(go_port, go_pin);
	output(led_direction, led_pin);
	output(digit_direction, digit_pin);
	output(go_direction, go_pin);
	output(serial_direction, serial_pin_out);
	//
	// init A/D
	//
	ADMUX = (1 << REFS1) | (0 << REFS0) // 1.1V ref
	  | (0 << MUX5) | (0 << MUX4) | (1 << MUX3)
	  | (1 << MUX2) | (0 << MUX1) | (1 << MUX0);  // 001101 = 20(PA1-PA2)
	ADCSRA = (1 << ADEN) // enable
	  | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // prescaler /128
	ADCSRB = (1 << BIN); // bipolar mode
	//
	// boot delay
	//
	set(led_port, led_pin);
	_delay_ms(500);
	clear(led_port, led_pin);
	//
	// main loop
	//
	while (1) {
		//
		// accumulate samples
		//
		negative = 0;
		accum = 0;
		for (count = 0; count < nsamples; ++count) { 
			//
			// initiate conversion
			//
			ADCSRA |= (1 << ADSC);
			//
			// wait for completion
			//
			while (ADCSRA & (1 << ADSC))
			;
			//
			// add result
			//
			if (ADCH & (1 << 1)) {
				// negative
				negative += 1;
				accum += (~(ADC) & 1023) + 1;
			} else {
				// positive
				negative -= 1;
				accum += ADC;
			}
		}
		if (negative > 0) {
			negative = 1;
		}
		else negative = -1;

		//
		// send result to serial probe
		//
		put_char(&serial_port, serial_pin_out, negative);
		char_delay();
		put_char(&serial_port, serial_pin_out, ((accum >> 8) & 255));
		char_delay();

		if (((accum >> 8) & 255) >= threshold) {
			if (negative == 1) {
				set(go_port, go_pin);
				clear(digit_port, digit_pin);
				set(led_port, led_pin);
			}
			else if (negative == -1) {
				clear(led_port, led_pin);
				set(digit_port, digit_pin);
				clear(go_port, go_pin);
			}
			else {
				set(go_port, go_pin);
				set(digit_port, digit_pin);
				clear(led_port, led_pin);
			}
			_delay_ms(300);
		}
		else {
			set(go_port, go_pin);
			set(digit_port, digit_pin);
			clear(led_port, led_pin);
		}
	}
}
